/* Implement a X Window System aplication window with C++ and
enable set event handlers with C.

FEATURES:
create one app window, init app, set palette, close window,
update the window, set the event loop

Works under: 8, 15, 16, 24, 32 BPP (not very tested)

User can draw on one frame buffer (called user frame buffer) as it draw on
video memory on standard vga modes. The user frame buffer is accessible with
the m_pyVidPal member variable. Example, to set one pixel on (10, 20) if
user frame buffer has size 320x200:
 *( m_pyVidPal + 320 * 20 + 10 ) = pixel_color

TIPS:
- the reference system is the same as the video:   (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX
- to create one X Window app follow the steps:
    create the app with mCreate;
    set the event loop calling mExecApp;
    the app exit when user press ESC.
    User can make something in the UserHandler function.
- the macro COMP_SHAREMEMORY - define the use of shared memory, speeding up the output.
- Palette emulation:
  If X display is using more than 8BPP, the palette mode is emulated.
  With this, user draw on an 8BPP frame buffer (user frame buffer) and the program
  convert this frame buffer to the display frame buffer - with the correct bit-deph -
  automaticaly.
  For the emulation is created one virtual palette and the user receive one
  8BPP frame buffer to draw (user frame buffer). Before show, the user frame buffer is
  copyed for the display frame buffer using the virtual palette RGBs to know the colors.
  The display frame buffer is the aplication window frame buffer, and has the current 
  X Window bits per pixel.
  
WARNING:
- the X type KeySym must be castable to DWORD
*/



#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#define XK_MISCELLANY 1
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysymdef.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>

#ifdef COMP_SHAREMEMORY
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif

#include "defs.h"
#include "keyboard.h"
#include "xwinapp.h"



extern CXWindowApp oAPP;



// data to the X system
static XrmOptionDescRec astTABLE [] = {
   { "-flag",   "Flag",                XrmoptionNoArg,  (caddr_t) "on" },
   { "-param",  "parameter",           XrmoptionSepArg, NULL },
};

typedef struct {
   int nFlag;
   int nParameter;
   char *pyGeometry;
   int nBorderwidth;
} pAppResourceRec, *pAppRes;

static pAppResourceRec stAPP_RESOURCES;
static char szWIN_DIMENSION [ 100 ];

static XtResource astRESOURCES [] = {
  { "param", "Parameter", XtRInt, sizeof ( int ),
    XtOffset ( pAppRes, nParameter ), XtRImmediate, (caddr_t) 80},
  { "flag", "Flag", XtRBoolean, sizeof ( Boolean ),
    XtOffset ( pAppRes, nFlag ), XtRImmediate, (caddr_t) TRUE },
  { XtNgeometry,  "Geometry", XtRString, sizeof ( caddr_t ),
    XtOffset ( pAppRes, pyGeometry ), XtRString, (caddr_t) szWIN_DIMENSION },
  { XtNborderWidth, "borderWidth", XtRInt, sizeof ( int ),
    XtOffset ( pAppRes, nBorderwidth ), XtRImmediate, (caddr_t) 1},
};



// the message handlers
void WindowHandler (  Widget stWi, XtPointer stClosure, XEvent* pstEvent,
                      Boolean* bCont_to_dispatch );



/* constructor
**********************************************************/
CXWindowApp::CXWindowApp ()
{
   m_pstDisplay = NULL;
   m_stTopLevel = 0;
   m_stShell = 0;
   m_stBox = 0;

   // window data
   m_nScreenNum = 0;
   m_nDpyWidth = 0;
   m_nDpyHeight = 0;
   m_pXImg = NULL;
   m_pyVidPal = m_pyVid = NULL;

   m_nBpp = 0;

   m_stColorMap = 0;
   szWIN_DIMENSION [ 0 ] = 0;
   m_dwResX = m_dwResY = 0;
}



/* destructor
**********************************************************/
CXWindowApp::~CXWindowApp ()
{
}



/* Create a window, init event handler,...
**********************************************************/
void CXWindowApp::mCreate ( char* szWinTitle, DWORD dwResX, DWORD dwResY,
                            int nArgc, char* pszArgv [] )
{
   // define the resolution
   m_dwResX = dwResX;
   m_dwResY = dwResY;
   sprintf ( szWIN_DIMENSION, "%dx%d", m_dwResX, m_dwResY );

   // init the app and window

   m_stTopLevel = XtVaAppInitialize ( &m_stApContext, szWinTitle,   
                  astTABLE, XtNumber ( astTABLE ), &nArgc, pszArgv,
                  NULL, NULL, 0 );

   XtGetApplicationResources ( m_stTopLevel, &stAPP_RESOURCES, astRESOURCES,
                               XtNumber ( astRESOURCES ), NULL, (Cardinal) 0 );

   //XParseGeometry (stAPP_RESOURCES.geometry, &geom_x, &geom_y, &geom_width,
   //&geom_height);

   m_stShell = XtVaCreatePopupShell ( szWinTitle,
               topLevelShellWidgetClass,
               m_stTopLevel,
               XtNgeometry, stAPP_RESOURCES.pyGeometry,
               XtNborderWidth, stAPP_RESOURCES.nBorderwidth,
               0 );

   //stBox = XtCreateManagedWidget ("box", boxWidgetClass, stShell, NULL, (Cardinal) 0);
   m_stBox = XtCreateWidget ( "box", boxWidgetClass, m_stShell, NULL, (Cardinal) 0 );

   XtPopup ( m_stShell, XtGrabNone );
 
   // get some new window data
   Screen*
      pScreenPtr;
   m_pstDisplay = XtDisplay ( m_stShell );
   m_nScreenNum = DefaultScreen ( m_pstDisplay );
   pScreenPtr   = DefaultScreenOfDisplay ( m_pstDisplay );
   m_nDpyWidth  = DisplayWidth ( m_pstDisplay, m_nScreenNum );
   m_nDpyHeight = DisplayHeight ( m_pstDisplay, m_nScreenNum );
   m_nBpp       = DefaultDepth ( m_pstDisplay, m_nScreenNum );
   m_stWindow   = XtWindow ( m_stShell );

   // configure drawing attributes
   XGCValues
      stValues;
   m_stGC = XCreateGC ( m_pstDisplay, m_stWindow, 0, &stValues );
   XSetForeground ( m_pstDisplay, m_stGC, BlackPixel ( m_pstDisplay, m_nScreenNum ) );
   XSetLineAttributes ( m_pstDisplay, m_stGC, 0, LineSolid, CapButt, JoinMiter );
   XSetGraphicsExposures ( m_pstDisplay, m_stGC, False );

   // create palette
   if ( m_nBpp == 8 )
      m_stColorMap = XCreateColormap ( m_pstDisplay, RootWindow ( m_pstDisplay, m_nScreenNum ), DefaultVisual ( m_pstDisplay, m_nScreenNum ), AllocAll );

   // set window message handler
   XtAddEventHandler ( m_stShell,
                       ExposureMask | KeyPressMask | KeyReleaseMask |
                       KeymapStateMask | FocusChangeMask |
                       PointerMotionMask | ButtonPressMask | ButtonReleaseMask, True,
                       WindowHandler, NULL );

   XMapWindow ( m_pstDisplay, XtWindow ( m_stShell ) );

   // set palette
   if ( m_nBpp == 8 )
      XSetWindowColormap ( m_pstDisplay, XtWindow ( m_stShell ), m_stColorMap );
}



/* Close the window, quit the app
**********************************************************/
void CXWindowApp::mDestroy ()
{
   if ( m_stShell )
     XtPopdown ( m_stShell );
   if ( m_stTopLevel )
     XtDestroyWidget ( m_stTopLevel );

   m_pyVid = NULL;
   if ( m_nBpp != 8 ) {     // desalocate indexed user frame buffer
      delete m_pyVidPal;
      m_pyVidPal = NULL;
   }   

   // bug: if end using exit the auto repeat goes NOT on!!!?? Why?
   //exit ( 0 );
   
   // enable autorepeat
   if ( m_pstDisplay )
      XAutoRepeatOn ( m_pstDisplay );
   
   printf ( "\n\nHey!   Press Control-C to exit!!!\n" );
}



/* Write a memory array of RGB data in palette.
payMem - one palette, format: 3 bytes (RGB), 256 times (256 colors)
**********************************************************/
void CXWindowApp::mWritePal ( BYTE* payMem )
{
   // seek all palette writing RGB
   for ( DWORD dwcCol = 0; dwcCol < 768; dwcCol += 3 ) {

      // define RGB

      DWORD
         dwRed   = payMem [ dwcCol + 0 ] * 4,
         dwGreen = payMem [ dwcCol + 1 ] * 4,
         dwBlue  = payMem [ dwcCol + 2 ] * 4;
         
      // set RGB

      // if running at 8BPP, set X palette
      if ( m_nBpp == 8 ) {
 
         // ...* 256 to convert for X convensions (16 bits?)
         XColor
            stColor;
         stColor.pixel = dwcCol / 3;
         stColor.red   = dwRed   * 256;
         stColor.green = dwGreen * 256;
         stColor.blue  = dwBlue  * 256;
         stColor.flags = DoRed | DoGreen | DoBlue;
         XStoreColor ( m_pstDisplay, m_stColorMap, &stColor );
      }
      else {       // set emulation palette
      
         tagRGB
            *pstPalEmul = m_astPalEmul + dwcCol / 3;
         pstPalEmul->dwR = dwRed;
         pstPalEmul->dwG = dwGreen;
         pstPalEmul->dwB = dwBlue;
      }   
   }
}



/* Show a frame buffer on window area.
If emulating 8BPP mode, make conversion from 8BPP -> nBPP;
n is current X Window bits per pixel.
**********************************************************/
void CXWindowApp::mShow ()
{
   // if emulating palette mode (X deph is more than 8),
   // need translate the current user frame buffer to display frame buffer
   if ( m_nBpp != 8 ) {

      // current pixel to convert
      BYTE
         *pyPixelPal =  m_pyVidPal,        // the 8BPP (for emulation) frame buffer
         *pyEndPal = m_pyVidPal + m_dwResX * m_dwResY - 1, // end of the 8BPP buffer
         *pyPixelVid = m_pyVid;            // the X display frame buffer
         
      // scan user frame buffer   
      while ( pyPixelPal != pyEndPal ) {
         
         // set X Window image
         switch ( m_nBpp ) {
         
            case 15:
            case 16:
               *( (unsigned short int*) pyPixelVid ) = (unsigned short int) mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 2;
               break;
         
            case 32:
               *( (DWORD*) pyPixelVid ) = mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 4;
               break;
         
            case 24:
               *( (DWORD*) pyPixelVid ) = mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 3;
               break;
         }     
         
         pyPixelPal ++;
      }   
   }   
   
   #ifdef COMP_SHAREMEMORY
   XShmPutImage ( m_pstDisplay, m_stWindow, m_stGC, &m_stShmImg, 0, 0, 0, 0, m_dwResX, m_dwResY, False );
   #else
   XPutImage ( m_pstDisplay, m_stWindow, m_stGC, m_pXImg, 0, 0, 0, 0, m_dwResX, m_dwResY );
   #endif
}



/* Receive one palette color and return the RGB color, using
the X Window BPP.
At 8BPP just return the given index color.
dwPalCol - the indexed palette color
return the formated color
**********************************************************/
DWORD CXWindowApp::mFormatPixelColor ( DWORD dwPalCol )
{
   // if X Window at 8BPP
   if ( m_nBpp == 8 )
      return dwPalCol;
      
   // if W bpp is more than 8   

   // set color acording current X Window BPP
   tagRGB
      *pstPal = m_astPalEmul + dwPalCol;   // point the RGB for the indexed color

   // return the RGB on the correct format
   switch ( m_nBpp ) {

      case 16:             // 16 bpp
         {
            tagFormat16
               unRet;
            // clear
            unRet.nValue = 0;   
            // set 
            unRet.stBits.bitsR = ( pstPal->dwR & 0xF8 ) >> 3;
            unRet.stBits.bitsG = ( pstPal->dwG & 0xFC ) >> 2;
            unRet.stBits.bitsB = ( pstPal->dwB & 0xF8 ) >> 3;
            // return
            return unRet.nValue;
         }   
         break;

      case 15:             // 15 bpp
         {
            tagFormat15
               unRet;
            // clear
            unRet.nValue = 0;   
            // set 
            unRet.stBits.bitsR = ( pstPal->dwR & 0xF8 ) >> 3;
            unRet.stBits.bitsG = ( pstPal->dwG & 0xF8 ) >> 3;
            unRet.stBits.bitsB = ( pstPal->dwB & 0xF8 ) >> 3;
            // return
            return unRet.nValue;
         }   
         break;
 
      default:            // 24, 32 bpp
         return pstPal->dwB + ( pstPal->dwR << 8 ) + ( pstPal->dwB << 16 );
         break;
   }
 
   return 0;
}



/* This is the window event handler - receive input from the window
**********************************************************/
void WindowHandler (  Widget stWi, XtPointer stClosure, XEvent* pstEvent,
                      Boolean* bCont_to_dispatch )
{
   // disable GCC compiling warnings
   (void)stWi; (void)stClosure; (void)bCont_to_dispatch;

   switch ( pstEvent->type ) {

      case Expose:
         
         // set output buffer
         if ( ! oAPP.m_pXImg ) {

            oAPP.m_pXImg = XGetImage ( oAPP.m_pstDisplay, oAPP.m_stWindow, 0, 0, oAPP.m_dwResX, oAPP.m_dwResY, AllPlanes, ZPixmap );

            #ifdef COMP_SHAREMEMORY

            oAPP.m_stShmImg = *oAPP.m_pXImg;
            oAPP.m_stShmi.shmid = shmget ( IPC_PRIVATE, oAPP.m_pXImg->bytes_per_line * oAPP.m_pXImg->height, IPC_CREAT | 0777 );
            oAPP.m_stShmi.shmaddr = (char*) shmat ( oAPP.m_stShmi.shmid, 0, 0 );
            oAPP.m_stShmi.readOnly = FALSE;
            XShmAttach ( oAPP.m_pstDisplay, &oAPP.m_stShmi );
            oAPP.m_stShmImg.data = oAPP.m_stShmi.shmaddr;
            oAPP.m_pyVid = (unsigned char*) oAPP.m_stShmi.shmaddr;
            oAPP.m_stShmImg.obdata = (char*) &oAPP.m_stShmi;

            #else

            oAPP.m_pyVid = (unsigned char*) ( oAPP.m_pXImg->data );

            #endif
         }
         
         // set user frame buffer
         if ( oAPP.m_nBpp != 8 )
            // if emulating 8BPP, alloc the indexed frame buffer
            oAPP.m_pyVidPal = new BYTE [ oAPP.m_dwResX * oAPP.m_dwResY ];
         else
            // if at 8BPP, no emulation need - indexed frame buffer is the display buffer
            oAPP.m_pyVidPal = oAPP.m_pyVid;   
            
         // clear user frame buffer
         memset ( oAPP.m_pyVidPal, 0, oAPP.m_dwResX * oAPP.m_dwResY );

         break;

      case FocusIn:
         // disable keyboard autorepeat (to avoid multiple kb msgs for same key press)
         if ( oAPP.m_pstDisplay )
            XAutoRepeatOff ( oAPP.m_pstDisplay );
         break;

      case FocusOut:
         if ( oAPP.m_pstDisplay )
            XAutoRepeatOn ( oAPP.m_pstDisplay );
         break;

      case KeyPress:
         {
            KeySym
               stKey;
            int
               nKey2,
               nState;
 
            stKey = XLookupKeysym ( (XKeyEvent*) pstEvent, 0 );
            nState = ( (XKeyEvent*) pstEvent )->state;
            nKey2 = stKey;
            
            // set the key in key table
            oAPP.m_oKb.mSetKeyTable ( stKey, TRUE );
            
            //printf("key press: %d    %d %d\n",stKey, nKey2, nState);
 
            // X specific action
            //switch ( stKey ) {
            //}
         }
         break;
 
      case KeyRelease:
         {
 
            KeySym
               stKey;
            //int
              // nKey2,
              // nState;

            stKey = XLookupKeysym ( (XKeyEvent*) pstEvent, 0 );
            
            // reset the key in key table
            oAPP.m_oKb.mSetKeyTable ( stKey, FALSE );
 
            // X specific action
            switch ( stKey ) {
 
               case XK_Escape:
                   oAPP.mDestroy ();
                   break;

               default:
                  break;
            }
         }
         break;

      default:
         break;
   }
}
